""" Basic linear algebra for GVars. """

# Created by G. Peter Lepage (Cornell University) on 2014-04-27.
# Copyright (c) 2015 G. Peter Lepage. 
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version (see <http://www.gnu.org/licenses/>).
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import numpy 
import gvar 

def det(a):
    """ Determinant of matrix ``a``. 

    Args:
        a: Two-dimensional, square matrix/array of numbers 
            and/or :class:`gvar.GVar`\s.

    Returns:
        Deterimant of the matrix.

    Raises:
        ValueError: If matrix is not square and two-dimensional.
    """
    amean = gvar.mean(a)
    if amean.ndim != 2 or amean.shape[0] != amean.shape[1]:
        raise ValueError('Bad matrix shape: ' + str(a.shape))
    da = a - amean
    ainv = inv(amean)
    return numpy.linalg.det(amean) * (1 + numpy.matrix.trace(da.dot(ainv)))

def slogdet(a):
    """ Sign and logarithm of determinant of matrix ``a``. 

    Args:
        a: Two-dimensional, square matrix/array of numbers 
            and/or :class:`gvar.GVar`\s.

    Returns:
        Tuple ``(s, logdet)`` where the determinant of matrix ``a`` is
            ``s * exp(logdet)``.

    Raises:
        ValueError: If matrix is not square and two-dimensional.
    """
    amean = gvar.mean(a)
    if amean.ndim != 2 or amean.shape[0] != amean.shape[1]:
        raise ValueError('Bad matrix shape: ' + str(a.shape))
    da = a - amean
    ainv = inv(amean)
    s, ldet = numpy.linalg.slogdet(amean)
    ldet += numpy.matrix.trace(da.dot(ainv))
    return s, ldet

def eigvalsh(a, eigvec=False):
    """ Eigenvalues of Hermitian matrix ``a``.

    Args:
        a: Two-dimensional, square matrix/array of numbers 
            and/or :class:`gvar.GVar`\s.
        eigvec (bool): If ``True``, method returns a tuple of arrays
            ``(val, vec)`` where the ``val[i]`` are the
            eigenvalues. Arrays ``vec[:, i]`` are the corresponding
            eigenvectors of ``a`` when one ignores uncertainties (that is, 
            they are eigenvectors of ``gvar.mean(a)``). Only ``val`` is 
            returned if ``eigvec=False`` (default).

    Returns: 
        Array of eigenvalues of matrix ``a`` if parameter 
        ``eigvec==False`` (default).  where the ``val[i]`` are the
        eigenvalues; otherwise it returns a tuple of arrays ``(val, vec)`` 
        where the ``val[i]`` are the eigenvalues. Arrays ``vec[:, i]`` are 
        the corresponding eigenvectors of ``a`` when one ignores 
        uncertainties (that is, they are eigenvectors of ``gvar.mean(a)``).

    Raises:
        ValueError: If matrix is not square and two-dimensional.
    """
    amean = gvar.mean(a)
    if amean.ndim != 2 or amean.shape[0] != amean.shape[1]:
        raise ValueError('Bad matrix shape: ' + str(a.shape))
    da = a - amean 
    val, vec = numpy.linalg.eigh(amean)
    val = val + [vec[:, i].dot(da.dot(vec[:, i])) for i in range(vec.shape[1])]
    return (val, vec) if eigvec else val

def inv(a):
    """ Inverse of matrix ``a``. 

    Args:
        a: Two-dimensional, square matrix/array of numbers 
            and/or :class:`gvar.GVar`\s.

    Returns:
        The inverse of matrix ``a``.

    Raises:
        ValueError: If matrix is not square and two-dimensional.
    """
    amean = gvar.mean(a)
    if amean.ndim != 2 or amean.shape[0] != amean.shape[1]:
        raise ValueError('Bad matrix shape: ' + str(a.shape))
    da = a - amean
    ainv = numpy.linalg.inv(amean)
    return ainv - ainv.dot(da.dot(ainv))

def solve(a, b):
    """ Find ``x`` such that ``a.dot(x) = b`` for matrix ``a``. 

    Args:
        a: Two-dimensional, square matrix/array of numbers 
            and/or :class:`gvar.GVar`\s.
        b: One-dimensional vector/array of numbers and/or 
            :class:`gvar.GVar`\s, or an array of such vectors.
            Requires ``b.shape[0] == a.shape[1]``.

    Returns:
        The solution ``x`` of ``a.dot(x) = b``, which is equivalent
        to ``inv(a).dot(b)``.

    Raises:
        ValueError: If ``a`` is not square and two-dimensional.
        ValueError: If shape of ``b`` does not match that of ``a``
            (that is ``b.shape[0] != a.shape[1]``).
    """
    amean = gvar.mean(a)
    if amean.ndim != 2 or amean.shape[0] != amean.shape[1]:
        raise ValueError('Bad matrix shape: ' + str(a.shape))
    bmean = gvar.mean(b)
    if bmean.shape[0] != a.shape[1]:
        raise ValueError(
            'Mismatch between shapes of a and b: {} {}'.format(a.shape, b.shape)
            )
    xmean = numpy.linalg.solve(amean, bmean)
    ainv = inv(a)
    return xmean + ainv.dot(b-bmean - (a-amean).dot(xmean))



"""
Created by G. Peter Lepage (Cornell University) on 9/2011.
Copyright (c) 2011-18 G. Peter Lepage.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version (see <http://www.gnu.org/licenses/>).

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
"""

GVAR_VERSION = '8.3.6'

from distutils.core import setup
from distutils.extension import Extension
from distutils.command.build_ext import build_ext as _build_ext

# compile from existing .c files if USE_CYTHON is False
USE_CYTHON = False # True

class build_ext(_build_ext):
    # delays using numpy and cython until they are installed;
    # cython is optional (set USE_CYTHON)
    # this code adapted from https://github.com/pandas-dev/pandas setup.py
    def build_extensions(self):
        import numpy
        if USE_CYTHON:
            from Cython.Build import cythonize
            self.extensions = cythonize(self.extensions)
        numpy_include = numpy.get_include()
        for ext in self.extensions:
            ext.include_dirs.append(numpy_include)
        _build_ext.build_extensions(self)


# create gvar/_version.py so gvar knows its version number
with open("src/gvar/_version.py","w") as version_file:
    version_file.write(
        "# File created by lsqfit setup.py\nversion = '%s'\n"
        % GVAR_VERSION
        )

# extension modules
# Add explicit directories to the ..._dirs variables if
# the build process has trouble finding the gsl library
# or the numpy headers. This should not be necessary if
# gsl and numpy are installed in standard locations.
ext_args = dict(
    include_dirs=[],
    library_dirs=[],
    runtime_library_dirs=[],
    extra_link_args=[]
    )

ext = '.pyx' if USE_CYTHON else '.c'

ext_modules = [
    Extension("gvar._gvarcore", ["src/gvar/_gvarcore" + ext], **ext_args),
    Extension( "gvar._svec_smat", ["src/gvar/_svec_smat" + ext], **ext_args),
    Extension("gvar._utilities", ["src/gvar/_utilities" + ext], **ext_args),
    Extension("gvar.dataset", ["src/gvar/dataset" + ext], **ext_args),
    Extension("gvar._bufferdict", ["src/gvar/_bufferdict" + ext], **ext_args),
    ]

# packages
packages = ["gvar"]
package_dir = dict(gvar="src/gvar")
package_data = dict(gvar=['../gvar.pxd', '_svec_smat.pxd', '_gvarcore.pxd'])

setup(name='gvar',
    version=GVAR_VERSION,
    description='Utilities for manipulating correlated Gaussian random variables.',
    author='G. Peter Lepage',
    author_email='g.p.lepage@cornell.edu',
    cmdclass={'build_ext':build_ext},
    packages=packages,
    package_dir=package_dir,
    package_data=package_data,
    ext_modules= ext_modules,
    # for pip (distutils ignores):
    install_requires=['cython>=0.17', 'numpy>=1.7'] if USE_CYTHON else ['numpy>=1.7'],
    # for distutils (pip ignores):
    requires=['cython (>=0.17)', 'numpy (>=1.7)'] if USE_CYTHON else ['numpy (>=1.7)'],
    url="https://github.com/gplepage/gvar.git",
    license='GPLv3+',
    platforms='Any',
    long_description="""\
    This package facilitates the creation and manipulation of arbitrarily
    complicated (correlated) multi-dimensional Gaussian random variables.
    The random variables are represented by a new data type that can be used
    in arithmetic expressions and pure Python functions. Such
    expressions/functions create new Gaussian random variables
    while automatically tracking statistical correlations between the new
    and old variables. This data type is useful for simple error propagation
    but also is heavily used by the Bayesian least-squares fitting module
    lsqfit.py (to define priors and specify fit results, while accounting
    for correlations between all variables).

    This package uses numpy for efficient array arithmetic, and cython to
    compile efficient core routines and interface code.
    """
    ,
    classifiers = [                     #
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Cython',
        'Topic :: Scientific/Engineering'
        ],
)